"""
Prepare JSON Response Data
"""
from flask import jsonify


def prepare_json_response(sms, success=True, data=None, code=200):
    """
    Prepare Json response
    """
    json_object = {'data': data} if data is not None else {}
    return jsonify({
        'message': sms,
        'success': success,
        **json_object
    }), code



"""
Binary UUID Class
"""
from abc import ABC
from uuid import UUID
from sqlalchemy.dialects.mysql import BINARY
from sqlalchemy.types import TypeDecorator


class BinaryUUID(TypeDecorator, ABC):
    """Optimize UUID keys. Store as 16 bit binary, retrieve as uuid.
    inspired by:
        https://mysqlserverteam.com/storing-uuid-values-in-mysql-tables/
    """

    impl = BINARY(16)
    cache_ok = True

    def process_bind_param(self, value, dialect):
        """
        Process create data
        """
        try:
            return value.bytes
        except AttributeError:
            try:
                return UUID(value).bytes
            except TypeError:
                # for some reason we ended up with the byte string
                # ¯\_(ツ)_/¯
                # I'm not sure why you would do that,
                # but here you go anyway.
                return value

    def process_result_value(self, value, dialect):
        """
        Process result data
        """
        return UUID(bytes=value)



"""
Model Utils
"""
import json
from abc import ABC

from sqlalchemy import TypeDecorator, Text, Integer

from .func_date import convert_date_to_timestamp, convert_timestamp_to_date


class JsonEncodeDict(TypeDecorator, ABC):
    """
    Enables JSON storage by encoding and decoding on the fly.
    """
    impl = Text

    def process_bind_param(self, value, dialect):
        """
        :param value:
        :param dialect:
        :return:
        """
        return '{}' if value is None else json.dumps(value)

    def process_result_value(self, value, dialect):
        """
        :param value:
        :param dialect:
        :return:
        """
        return {} if value is None else json.loads(value)


class TimestampField(TypeDecorator, ABC):
    """
    Enables JSON storage by encoding and decoding on the fly.
    """
    impl = Integer
    cache_ok = True

    def process_bind_param(self, value, dialect):
        """
        :param value:
        :param dialect:
        :return:
        """
        return convert_date_to_timestamp(value)

    def process_result_value(self, value, dialect):
        """
        :param value:
        :param dialect:
        :return:
        """
        return convert_timestamp_to_date(value)


def generic_get_serialize_data(schema_object, query):
    """
    Get Serialize data
    :param schema_object:
    :param query:
    :return:
    """
    return schema_object.dump(query)



"""
Pagination Generic Class
"""
from flask_marshmallow import Marshmallow


class PaginationSchema:
    """
    Pagination Class
    """
    schema_object = None
    ma = Marshmallow()

    def __init__(self, ma, app, schema_object):
        self.schema_object = schema_object
        self.ma = ma
        self.ma.init_app(app)
        self.pagination_sub_class = self.PaginationSchemaSubClass()

    class PaginationSchemaSubClass(ma.SQLAlchemyAutoSchema):
        """
        Pagination Schema Data
        """
        class Meta:
            """
            Meta Class
            """
            fields = ('has_prev', 'has_next', 'next_num', 'page', 'pages', 'per_page', 'prev_num', 'total', 'items')



"""
Validation utils
"""
from flask import request

from .prepare_json_response import prepare_json_response


def validate_generic_form(validation_class):
    """
    Generic validation form
    :param validation_class: Validation class
    :return: jsonfy | None
    """
    errors = validation_class.validate(request.json)
    if errors:
        return prepare_json_response('Validation Error', False, {'errors': errors}, 422)
    return None



"""
Function date methods
"""
from datetime import datetime
import pytz as pytz


def convert_date_to_timestamp(date, format_datetime='%Y-%m-%d %H:%M:%S'):
    """
    Convert date to timestamp
    """
    if isinstance(date, str):
        date = datetime.strptime(date, format_datetime)

    if not isinstance(date, datetime):
        raise ValueError('Incorrect data type')

    return int(date.astimezone(pytz.UTC).timestamp())


def convert_timestamp_to_date(timestamp):
    """
    Convert timestamp to date
    """
    return datetime.fromtimestamp(timestamp)


def get_timestamp_now():
    """
    Get timestamp now
    """
    return int(datetime.now().astimezone(pytz.utc).timestamp())



"""
Abstract Base Model
"""
from datetime import datetime

from sqlalchemy import Column, DateTime, Integer


class BaseModel:
    """
    Datetime created_at and updated_at
    """
    id = Column(Integer, primary_key=True, autoincrement=True)
    created_at = Column(DateTime, default=datetime.now())
    updated_at = Column(DateTime, default=datetime.now(), onupdate=datetime.now())



"""
Setup configuration file
"""
from setuptools import setup, find_packages

setup(
    name='ms_general_utils',
    packages=find_packages(),
    include_package_data=True,
    version='1.0.6',
    description='General functions for the implementation of microservices.',
    authors=[
        {"name": "Alejandro A. Serrano Correa", "email": "alejandroasc93@gmail.com"},
        {"name": "Rene Gonzalez Ramos", "email": "rgramos9310@gmail.com"}
    ],
    license="GPLv3",
    url="https://github.com/rgramos/ms-utils.git",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)"
    ],
)



"""
Generic Views
"""
from flask.views import MethodView
from ms_utils import ViewGeneralMethods

ViewGeneralMethods.db = db
vmg = ViewGeneralMethods()


class GenericItemCrud(MethodView):
    """
    Generic Crud Class
    """
    init_every_request = False
    db = None
    model = None
    validator = None
    schema = None

    def __init__(self, db, model, validator, schema):
        self.db = db
        self.model = model
        self.validator = validator
        self.schema = schema
        ViewGeneralMethods.db = db

    vmg = ViewGeneralMethods()

    def get(self, object_id):
        """
        Component Type details
        :param object_id:
        :return:
        """
        return vmg.generic_details(self.model, self.schema, object_id)

    def patch(self, object_id):
        """
        Component Type update
        :param object_id:
        :return:
        """
        return vmg.generic_update_or_create(self.model, self.validator, object_id)

    def delete(self, object_id):
        """
        Component Type update
        :param object_id:
        :return:
        """
        return vmg.generic_delete(self.model, object_id)


class GenericGroupCrud(MethodView):
    """
    Generic Crud Class
    """
    init_every_request = False
    model = None
    validator = None
    schema = None

    def __init__(self, db, model, validator, schema):
        self.db = db
        self.model = model
        self.validator = validator
        self.schema = schema
        ViewGeneralMethods.db = db

    vmg = ViewGeneralMethods()

    # Component Type
    def get(self):
        """
        Component type list
        :return: jsonify
        """
        return vmg.generic_list(self.model, self.schema)

    def post(self):
        """
        Create component type
        :return: jsonfy
        """
        return vmg.generic_update_or_create(self.model, self.validator)


def register_api(app, model, name, validator, schema):
    """
    Register API method
    :param schema:
    :param validator:
    :param app:
    :param model:
    :param name:
    :return:
    """
    item = GenericItemCrud.as_view(f"{name}-item", model, validator, schema)
    group = GenericGroupCrud.as_view(f"{name}-group", model, validator, schema)
    app.add_url_rule(f"/{name}/<int:id>", view_func=item)
    app.add_url_rule(f"/{name}/", view_func=group)



"""
Init imports
"""
from .binary_uuid import BinaryUUID
from .func_date import convert_date_to_timestamp, get_timestamp_now, convert_timestamp_to_date
from .prepare_json_response import prepare_json_response
from .generic_pagination import PaginationSchema
from .validation_utils import validate_generic_form
from .abstract_model import BaseModel
from .model_utils import JsonEncodeDict, generic_get_serialize_data, TimestampField
from .view_utils import ViewGeneralMethods
from .generic_crud_class import register_api, GenericItemCrud, GenericGroupCrud



"""
View Utils
"""
from flask_marshmallow import Marshmallow

from ms_utils import prepare_json_response, PaginationSchema
from flask import current_app, request

from .model_utils import generic_get_serialize_data
from .validation_utils import validate_generic_form


class ViewGeneralMethods:
    """
    View generic Methods
    """
    ma = Marshmallow()
    db = None

    def __int__(self, ma, app, db):
        self.db = db
        self.ma = ma
        self.ma.init_app(app)

    def generic_list(self, model, schema):
        """
        Generic list
        :return: jsonify
        """
        page = int(request.args.get('page')) if request.args.get('page') else 1
        per_page = int(request.args.get('per_page')) if request.args.get('per_page') else 10

        query = model.query.filter(model.name.contains(request.args.get('q'))) \
            .paginate(page=page, per_page=per_page) if request.args.get(
            'q') else model.query.paginate(page=page, per_page=per_page)
        query.items = generic_get_serialize_data(schema(many=True), query.items)
        data = generic_get_serialize_data(
            PaginationSchema(self.ma, current_app, schema(many=True)).pagination_sub_class, query)

        return prepare_json_response(f'{model.__name__} get successfully', data=data)

    def generic_update_or_create(self, model, validation_class, object_id=None):
        """
        Generic method for create or update provider
        :param validation_class:
        :param model:
        :param object_id:
        :return: jsonify
        """
        errors = validate_generic_form(validation_class)
        if errors is not None:
            return errors
        action_text = 'created'
        try:
            data = request.json
            if object_id is None:
                self.generic_create(model, data)
            else:
                action_text = 'updated'
                model_object = self.db.get_or_404(model, object_id)
                self.generic_update(model_object, data)
            self.db.session.commit()
        except ValueError:
            self.db.session.rollback()
            return prepare_json_response(f'{model.__name__} can not be {action_text} successfully', False, code=400)
        return prepare_json_response(f'{model.__name__} {action_text} successfully')

    def generic_create(self, model, data):
        """
        Generic Create method
        :param model:
        :param data:
        :return:
        """
        model_object = model(**data)
        self.db.session.add(model_object)

    @staticmethod
    def generic_update(model_object, data):
        """
        Generic Update method
        :param data:
        :param model_object:
        :return:
        """
        model_object.update_from_json(data)

    def generic_details(self, model, schema, object_id):
        """
        Generic details method
        :param schema:
        :param model:
        :param object_id:
        :return:
        """
        model_object = self.db.get_or_404(model, object_id)
        return prepare_json_response(f'{model.__name__} get successfully',
                                     data=generic_get_serialize_data(schema, model_object))

    def generic_delete(self, model, object_id):
        """
        Generic delete method
        :param model:
        :param object_id:
        :return:
        """
        model_object = self.db.get_or_404(model, object_id)
        model_object.delete()
        return prepare_json_response(f'{model.__name__} deleted successfully!')

    def generic_change_boolean(self, model, object_id, field):
        """
        Generic change boolean method
        :param model:
        :param object_id:
        :param field:
        :return:
        """
        model_object = self.db.get_or_404(model, object_id)
        model_object[field] = not model_object[field]
        self.db.session.commit()
